' mt19937-64.vb (for Visual Basic .NET), by Ron Charlton, based on
' C code that is Copyright (c) 2004 by Takuji Nishimura and Makoto
' Matsumoto.  mt19937-64.vb is Copyright (c) 2014 by Ron Charlton.
' The C source code comments below have detailed redistribution
' requirements and disclaimers that apply to mt19937-64.vb.
'
' This file implements the 64-bit Mersenne Twister (MT), mt19937-64,
' pseudorandom number generator (PRNG) as a Visual Basic .NET (VB)
' class.  Its period is 2^19937-1; approximately 10^6002.
' It is equidistributed in 311 dimensions.
'
' $RCSfile: mt19937-64.vb $
' $Revision: 1.8b $
' $Date: 2015-03-31 03:58:47-05 $
' $Name:  $
'
' TO USE 64-bit Mersenne Twister in your VB project:
'
'   1) Add only this file, mt19937-64.vb, to your VB project
'
'	2) Check "Remove Integer Overflow Checks" for your project in VB's
'	   SolutionExplorer-MyProject-Compile-AdvancedCompilerOptions...

'   3) mt19937-64.vb has a genrand64_int128SignedInt() function
'      that REQUIRES VB 2010 or later for data type BigInteger.
'      The function is disabled by default.  If you want to use
'	   genrand64_int128SignedInt() do the following:
'	   a) Add a project reference to BigInteger with
'		  Visual Basic menu item
'			[VB 2010] Project/AddReference/.NET/System.Numerics
'			[VB 2012] Project/AddReference/Assemblies/Framework/System.Numerics
'	   b) Set Need128bitInts to True below.
'
' Set to True if you need genrand64_int128SignedInt() with Visual
' Basic 2010 or later.  Set it to False otherwise (must be False
' for Visual Basic 2008 or earlier).
#Const Need128bitInts = False
' (Also set the #Const in Demo64.vb if you include that file in a
' project.)
'
' Important information follows in these header comments.  Please
' skim them if you have trouble with MTRandom64.
'
' To see a list of PRNGs herein, search for "FUNCTIONS:" (no quotes).
' To see instructions for using the PRNGs, search for "USE:" (no quotes).
' To see performance results, search for "PERFORMANCE:" (no quotes).
'
' Class MTRandom64 was implemented in two steps:
'
'	1)	A C-program for MT19937-64 (2004/9/29 version).  Coded by Takuji
'		Nishimura and Makoto Matsumoto.
'
'	2)	Translation from C (mt19937-64.c) to Visual Basic .NET was made
'		and tested by me, Ron Charlton (2014-02-18). I also translated
'		genrand64_intMax() from C++ code by Richard J. Wagner and Magnus
'		Jonsson.
'
'	Ron Charlton
'	9002 Balcor Circle
'	Knoxville, TN 37923-2301 USA
'	Email:		charltoncr @ RonCharlton.org (remove spaces)
'	Home Page:	http://RonCharlton.org/
'	Start Date:	2014-02-18
'
' Bug reports, improvements, corrections, complaints or kudos are welcomed.
'
' The C version, mt19937-64.c, was obtained via a link at
' http://www.math.sci.hiroshima-u.ac.jp/~m-mat/MT/emt.html on 2014-02-17.
'
' For cryptographically secure random number generation, search
' Visual Basic .NET Help for "RNGCryptoServiceProvider class". Or
' see http://www.math.sci.hiroshima-u.ac.jp/~m-mat/MT/efaq.html.
'
' I compared the Visual Basic .NET Main()'s output file,
' "mt19937-64VBTest.out.txt", to file "mt19937-64.out.txt" from
' the original MT authors' web site. The files are identical.
' (See "USE:" below for details.)
'
' I also modified the original mt19937-64test.c code and VB Main() to
' produce 100,000,000 random integers and 100,000,000 random reals.
' The C & VB genrand64_int64() outputs were identical.
'
' The genrand64_real2() VB output was different from the C output for
' 7 reals out of 100,000,000.  BUT THE DIFFERENCE IS AN ILLUSION
' caused by VB and C rounding differences when converting to
' a string.
'
' I compared 100,000,000 genrand64_real2() outputs for C and VB by
' using a VB union to show the VB genrand64_real2() output bit
' patterns in hexadecimal.  The 100,000,000 hexadecimal outputs for
' C and VB genrand64_real2() are identical.
'
' VB Format() rounds all Double outputs to no more than 15
' places, regardless of the Format() string. The 7 reals out of
' 100,000,000 where VB differed are explained by printing them
' to 14 decimal places:
'
'	C: 0.73880033999999  VB: 0.73880034000000
'	C: 0.58501159999999  VB: 0.58501160000000
'	C: 0.77837456999999  VB: 0.77837457000000
'	C: 0.90141489999999  VB: 0.90141490000000
'	C: 0.45203357999999  VB: 0.45203358000000
'	C: 0.51149137999999  VB: 0.51149138000000
'	C: 0.13164961999999  VB: 0.13164962000000
'
' Note: In Visual Basic 2012, 2010, 2008 and 2005 --
'	Data type Integer    has a range of -2^31 inclusive to 2^31-1 inclusive.
'	Data type UInteger   has a range of   0   inclusive to 2^32-1 inclusive.
'	Data type Long       has a range of -2^63 inclusive to 2^63-1 inclusive.
'	Data type ULong      has a range of   0   inclusive to 2^64-1 inclusive.
'   Data type BigInteger has an unlimited range and is signed


' Comments (with redistribution requirements and disclaimers) from
' the original MT authors' C source code (mt19337-64.c) follow.
'/*
'   A C-program for MT19937-64 (2004/9/29 version).
'   Coded by Takuji Nishimura and Makoto Matsumoto.
'
'   This is a 64-bit version of Mersenne Twister pseudorandom number
'   generator.
'
'   Before using, initialize the state by using init_genrand64(seed)  
'   or init_by_array64(init_key, key_length).
'
'   Copyright (C) 2004, Makoto Matsumoto and Takuji Nishimura,
'   All rights reserved.                          
'
'   Redistribution and use in source and binary forms, with or without
'   modification, are permitted provided that the following conditions
'   are met:
'
'     1. Redistributions of source code must retain the above copyright
'        notice, this list of conditions and the following disclaimer.
'
'     2. Redistributions in binary form must reproduce the above copyright
'        notice, this list of conditions and the following disclaimer in the
'        documentation and/or other materials provided with the distribution.
'
'     3. The names of its contributors may not be used to endorse or promote 
'        products derived from this software without specific prior written 
'        permission.
'
'   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
'   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
'   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
'   A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
'   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
'   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
'   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
'   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
'   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
'   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
'   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'
'   References:
'   T. Nishimura, ``Tables of 64-bit Mersenne Twisters''
'     ACM Transactions on Modeling and 
'     Computer Simulation 10. (2000) 348--357.
'   M. Matsumoto and T. Nishimura,
'     ``Mersenne Twister: a 623-dimensionally equidistributed
'       uniform pseudorandom number generator''
'     ACM Transactions on Modeling and 
'     Computer Simulation 8. (Jan. 1998) 3--30.
'
'   Any feedback is very welcome.
'   http://www.math.sci.hiroshima-u.ac.jp/~m-mat/MT/emt.html
'   email: m-mat @ math.sci.hiroshima-u.ac.jp (remove spaces)
'*/
' [End mt19937-64.c comments]

' The redistribution requirements and the disclaimers above
' apply to this file, mt19937-64.vb, and to its author as
' a contributor.
'
' It would be nice to CC: rjwagner@writeme.com and charltoncr @ RonCharlton.org
' when you write to the original authors.
'---
'_________________________________________________________________________________________
'FUNCTIONS:
'       Function                      Returns values in the range:
'       -------------------------     -------------------------------------------------------
'       genrand64_int64()             [0,18446744073709551615]  (0 to 2^64-1)
'
'       genrand64_int63()             [0,9223372036854775807]  (0 to 2^63-1)
'
'       genrand64_real1()             [0.0,1.0]   (both 0.0 and 1.0 included)
'
'       genrand64_real2()             [0.0,1.0)   (0.0 included, 1.0 excluded)
'
'       genrand64_real3()             (0.0,1.0)   (both 0.0 and 1.0 excluded)
'
'
'		The following three functions were added by Ron Charlton for Visual Basic .NET. 
'
'		genrand64_intMax(upper)			[0,upper] for upper < 2^64
'												(0 to 18446744073709551615 but <= upper)
'
'		genrand64_intRange(lower,upper)	[lower,upper] for 0 <= lower <= upper < 2^64
'												(0 <= lower <= upper <= 18446744073709551615)
'
'		Optional:
'       genrand64_int128SignedInt()     [-2^127,2^127-1]
'                                               (-170141183460469231731687303715884105728 to
'                                                 170141183460469231731687303715884105727)
'_________________________________________________________________________________________
'OPERATORS:
'		Operator  =(ByVal v As MTRandom64, ByVal w As MTRandom64) As Boolean
'		Operator <>(ByVal v As MTRandom64, ByVal w As MTRandom64) As Boolean
'_________________________________________________________________________________________
'SUBS:
'       Sub                       Argument
'       ------------------------  ---------------------------------------------------------
'       init_genrand64(seed)      any seed included in [0,18446744073709551615]
'       init_by_array64(array)    array has elements of type ULong;
'                                 the array must have at least one element
'
'		These four Subs were added for the Visual Basic .NET version. 
'
'       init_by_array32(array)    array has elements of type UInteger;
'                                 the array must have at least two elements
'		init_by_MTRandom64(r)	  r is another MTRandom64 instance
'       init_by_random64(boolean) True:  reseed VB Random from the system clock,
'										 then reseed MTRandom64 from VB Random
'                                 False: use the next two values from VB Random to
'										 reseed MTRandom64
'       init_by_crypto64(0.0)     0.0 (cryptographically randomize the
'                                 PRNG state; this DOES NOT make mt19937-64
'                                 cryptographically secure, it provides the
'                                 full range of possible initialized states/
'                                 sequences).  Changing the argument value
'                                 (not its type) will not affect initialization
'		saveState64("fileName")	  any valid file name (creates an XML file)
'		loadState64("fileName")	  name of any file saved earlier with saveState()
'
'
'_________________________________________________________________________________________
'USE:
' In your Visual Basic .NET application:
'
'	1) Add this file to your Visual Basic .NET project.
'
'	2) Remove Integer Overflow Checks (detailed at the top of this file).
'
'	3) Optional: For genrand64_int128SignedInt() with VB2010 and later, add
'	   a project reference to BigInteger with menu item
'			[VB 2010] Project/AddReference/.NET/System.Numerics
'			[VB 2012] Project/AddReference/Assemblies/Framework/System.Numerics
'      Set the #Const assignment near the first of this file as appropriate.
'
'	4) Create one or more instances of the MTRandom64 class.  The pseudorandom
'      sequence for each instance is initialized based on the arguments.
'	   Make certain each instance stays in scope for the duration of its use so
'	   it won't be re-initialized inadvertently.  Seven different methods of
'	   initialization are:
'
'			a) Dim r As New MTRandom64()
'					Initialize with seed == 5489 (the original MT
'					authors' default seed).
'
'			b) Dim r As New MTRandom64(19456UL)
'					Initialize with seed==19456.  Any unsigned long seed
'					in the range [1,18446744073709551615] is acceptable.
'
'			c) Dim init64() As ULong = {&H12345UL, &H23456UL, &H34567UL, &H45678UL}
'			   Dim r As New MTRandom64(init64)
'					Initialize with an array of ULong (as in the
'					original MT authors' version).  The array must have
'					at least one element. 312 elements are desirable; more
'					or fewer elements are acceptable.  The element values
'					ideally should be truly random for best results.
'
'			d) Dim r As New MTRandom64(True)
'					Seed VB Random from the system clock, and then
'					seed MTRandom64 with the next two values from VB Random.
'					True is required but ignored.
'
'			e) Dim r As New MTRandom64("filename.ext")
'					Initialize with a generator state saved earlier with
'					MTRandom64.saveState("filename.ext")
'
'           f) Dim r As New MTRandom64(0.0)
'                   Initialize with cryptographically random numbers from
'                   Visual Basic's RNGCryptoServiceProvider.  The argument
'                   0.0 is required but ignored.  Every new MTRandom64 instance
'                   that is initialized this way will have any one of
'                   approximately 2^19937 different beginning states.
'                   THIS DOES NOT MAKE MTRandom64 CRYPTOGRAPHICALLY SECURE.
'
'			g) Dim init32() As UInteger = {&H12345UL, &H23456UL, &H34567UL, &H45678UL}
'			   Dim r As New MTRandom64(init32)
'					Initialize with an array of UInteger.  The array must have
'					at least two elements.  624 elements are desirable; more
'					or fewer elements are acceptable.  The element values
'					ideally should be truly random for best results.
'
'			h) Dim p As New MTRandom64()
'			   Dim r As New MTRandom64(p)
'					Initialize PRNG r's state with PRNG p's state.
'
'	5) Call any of the genrand64_X() functions listed above.  You can re-
'      initialize an MTRandom64 instance at any point by calling init_genrand64(),
'	   init_by_array64(), init_by_random64(True), init_by_crypto64(), init_by_MTRandom64()
'	   or init_by_array32(). You can save or load the PRNG state at any point
'	   by calling saveState64() or loadState64().
'
'	6) Catch exceptions as desired.  The MTRandom64 exception classes are at the
'	   end of this file.
'
' To test genrand64_int64(), genrand64_real2(), init_genrand64() and init_by_array64()
' in the Visual Basic .NET version of 64-bit Mersenne Twister:
'
'	1) Create a Visual Basic .NET console application Project consisting of files
'			mt19937-64.vb
'			Demo64.vb
'			PerformanceTest64.vb
'			MakeEntropyTestData64.vb
'           MakeDiehardData64.vb
'			Main.vb
'
'	2) Follow the instructions at the top of this file.  Then build and run the project.
'
'	3) Compare the file "mt19937-64VBTest.out.txt" with the original MT authors'
'	   test output file ("mt19937-64.out.txt", found by following
'	   links at http://www.math.sci.hiroshima-u.ac.jp/~m-mat/MT/emt.html).  
'	   They should be identical.
'
'_________________________________________________________________________________________
'PERFORMANCE:
'	On a iBUYPOWER Pro Series P700 computer with 3.60 GHz Intel i7-3820 processor,
'	16 GB RAM, Visual Basic .NET (2012) compiler, and generating 100,000,000 random
'	numbers with each function (running the Release version at a Cmd.exe prompt)
'	on 64-bit Windows:
'
'									Seconds		Nanoseconds		Calls Per
'	Function						Run Time	Per Call		Second
'	--------------------------		--------	-----------		-----------
'	genrand64_int64()				 0.741		    7.41		135,000,000
'	genrand64_int63()				 0.787		    7.87		127,000,000
'	genrand64_real1()				 0.866			8.66		116,000,000
'	genrand64_real2()				 0.881			8.81		112,000,000
'	genrand64_real3()				 0.881			8.81		112,000,000
'	genrand64_intMax(1000)			 0.790		    7.90		 83,540,000
'	genrand64_intMax(&H5FFFFFFFUL)	 1.666		   16.66		 56,460,000
'	genrand64_intRange(10UL, 20UL)	 2.953		   29.53		 46,750,000
'   genrand64_int128SignedInt()		31.823		  318.23		  3,142,000
'
' NOTE: genrand64_int64() runs about 43 times faster than
' genrand64_int128SignedInt(). Data type BigInteger is SLOW on 32- and 64-bit
' Windows!
'_________________________________________________________________________________________
'=========================================================================================


Option Strict On
Option Explicit On

Imports System.Xml
Imports System.Xml.Serialization
Imports System.IO
Imports System.Security.Cryptography

' For the following import to work in VB 2010 and later, add a project reference:
' VB 2010 menu - Project/AddReference/.NET/System.Numerics
' VB 2012 menu - Project/AddReference/Assemblies/Framework/System.Numerics
#If Need128bitInts Then
Imports System.Numerics	' for type BigInteger used in genrand64_int128SignedInt()
#End If

' The 64-bit Mersenne Twister implemented as a VB class
Public Class MTRandom64

	' Constants from mt19937-64.c
	Private Const NN As Integer = 312
	Private Const MM As Integer = 156
	Private Const MATRIX_A As ULong = &HB5026F5AA96619E9UL
	Private Const UM As ULong = &HFFFFFFFF80000000UL ' Most significant 33 bits
	Private Const LM As ULong = &H7FFFFFFFUL		 ' Least significant 31 bits

#If Need128bitInts Then
	' Used by genrand64_int128SignedInt():
	Private int128SignedIntMaxValue, twoToThe128power As BigInteger
#End If

	' A class (within the MTRandom64 class) to hold the MT generator state
	<Serializable> Public Class MTState64
		Public mti As Integer
		Public mt(NN - 1) As ULong

		Sub New()
			Me.mti = NN + 1
		End Sub
	End Class

	' Here lies the state of a generator so it can be saved to an XML file
	' and loaded later.
	Public state64 As New MTState64()

	' a VB Random Class PRNG used in init_random() to seed MTRandom64 based
	' on the system clock
	Private rng As Random = Nothing

	Private mag01() As ULong = {0UL, MATRIX_A}


	' <><><><><><><><><><>[MTRandom64 Contructors]<><><><><><><><><><><>

	'' Instantiate an MTRandom64 PRNG and initialize from another MTRandom64 instance's state.
	Public Sub New(ByRef v As MTRandom64)
		Me.init_by_MTRandom64(v)
		Me.init_int128SignedInt()	  ' initialize constants for genrand64_int128SignedInt()
	End Sub


	'' Instantiate an MTRandom64 PRNG with the default seed
	Public Sub New()
		' The original authors' seed is used.
		Me.init_genrand64(5489UL)
		Me.init_int128SignedInt()	  ' initialize constants for genrand64_int128SignedInt()
	End Sub


	' Instantiate an MTRandom64 PRNG with a ULong seed
	Public Sub New(ByVal seed As ULong)
		Me.init_genrand64(seed)
		Me.init_int128SignedInt()	  ' initialize constants for genrand64_int128SignedInt()
	End Sub


	' Instantiate an MTRandom64 PRNG with a ULong array
	Public Sub New(ByRef array() As ULong)
		Me.init_by_array64(array)
		Me.init_int128SignedInt()	  ' initialize constants for genrand64_int128SignedInt()
	End Sub


	' Instantiate an MTRandom64 PRNG with a UInteger array
	Public Sub New(ByRef array() As UInteger)
		Me.init_by_array32(array)
		Me.init_int128SignedInt()	  ' initialize constants for genrand64_int128SignedInt()
	End Sub


	' Instantiate an MTRandom64 PRNG with a pseudorandom seed.  Variable dummy
	' is used only to distinguish this overload from the others.
	Public Sub New(ByVal dummy As Boolean)
		Me.init_by_Random64(True)
		Me.init_int128SignedInt()	  ' initialize constants for genrand64_int128SignedInt()
	End Sub


	' Instantiate an MTRandom64 PRNG with a cryptographically secure set of numbers.
	' THIS DOES NOT MAKE MTRandom64 CRYPTOGRAPHICALLY SECURE. It provides a good
	' set of initializers for general pseudorandom number use.
	Public Sub New(ByVal dummy As Double)
		Me.init_by_crypto64(0.0)
		Me.init_int128SignedInt()	  ' initialize constants for genrand64_int128SignedInt()
	End Sub


	' Instantiate an MTRandom64 PRNG with an XML file created by MTRandom64.saveState()
	Public Sub New(ByVal fileName As String)
		Me.loadState64(fileName)
		Me.init_int128SignedInt()	  ' initialize constants for genrand64_int128SignedInt()
	End Sub
	' <><><><><><><><><><>[End of MTRandom64 Contructors]<><><><><><><><><><><>



	' Initialize with another MTRandom64 instance.
	Public Sub init_by_MTRandom64(ByRef v As MTRandom64)
		' Copy v's state to this MTRandom64 instance's state.
		Array.Copy(v.state64.mt, state64.mt, state64.mt.Length)
		state64.mti = v.state64.mti
	End Sub


	' Initialize with a 32-bit, unsigned array.
	Public Sub init_by_array32(ByRef init_key() As UInteger)
		If init_key.Length < 2 Then
			Throw New ArgumentException("init_by_array32() argument ""array"" must have at least 2 elements.")
		End If
		Dim ik64((init_key.Length \ 2) - 1) As ULong
		Dim idx As Integer
		For idx = 0 To ik64.GetUpperBound(0)
			ik64(idx) = CULng(init_key(idx * 2)) Or (CULng(init_key(idx * 2 + 1)) << 32)
		Next
		Me.init_by_array64(ik64)
	End Sub


	' Initialize the MTRandom64 PRNG with a pseudorandom seed from VB's Random class.
	' reSeedFromClock: True  - reseed the INITIALIZER RNG from the system clock
	'                  False - use the next 2 numbers from VB Random variable rng
	Public Sub init_by_Random64(ByVal reSeedFromClock As Boolean)

		' If this is the first call of init_Random64() or user asks for reseed
		' of rng from the system clock.
		' (Must check for rng Is Nothing because user might re-initialize
		' the MTRandom64 instance when it was first initialized another way.)
		If rng Is Nothing Or reSeedFromClock Then
			' seed rng from the system clock by making a new instance
			rng = New Random()
		End If

		' initialize MTRandom64 with 2 pseudorandom Integers from rng
		Dim ui1 As UInteger = CUInt(CLng(rng.Next(Int32.MinValue, Int32.MaxValue)) - CLng(Int32.MinValue))
		Dim ui2 As UInteger = CUInt(CLng(rng.Next(Int32.MinValue, Int32.MaxValue)) - CLng(Int32.MinValue))
		Me.init_genrand64((CULng(ui1) << 32) Or CULng(ui2))
	End Sub


	' Initialize the MTRandom64 PRNG with a cryptographically secure (?) set
	' of numbers. THIS DOES NOT MAKE MTRandom64 CRYPTOGRAPHICALLY SECURE.
	' It provides the full range of approximately 2^19937 possible initial
	' states/sequences.
	Public Sub init_by_crypto64(ByVal dummy As Double)
		' I tried to use a union [see 'unions (VB)' in Help Index] but a
		' byte array is not an object and cannot be overlapped with an
		' object (ULong).  Error occurs at runtime.
		Const bytesPerULong As Integer = 8
		Dim rngCsp As New RNGCryptoServiceProvider()
		' Create a byte array to hold the random values.
		Dim bytes(bytesPerULong * state64.mt.Length) As Byte
		' Fill the array with random values.
		rngCsp.GetBytes(bytes)

		' Fill state64.mt() with the pseudorandom values in bytes().
		Dim i As Integer
		For i = 0 To state64.mt.GetUpperBound(0)
			Dim baseByte As Integer = i * bytesPerULong
			state64.mt(i) = _
				CULng(bytes(baseByte + 7)) << 56 Or _
				CULng(bytes(baseByte + 6)) << 48 Or _
				CULng(bytes(baseByte + 5)) << 40 Or _
				CULng(bytes(baseByte + 4)) << 32 Or _
				CULng(bytes(baseByte + 3)) << 24 Or _
				CULng(bytes(baseByte + 2)) << 16 Or _
				CULng(bytes(baseByte + 1)) << 8 Or _
				CULng(bytes(baseByte + 0))
		Next i
		' Make SURE the state isn't all zeroes.
		' "mt19937-64.c's method is silly.  It overwrites
		' 64 bits instead of setting one bit.
		state64.mt(0) = state64.mt(0) Or (1UL << 63)
		' Use the crypto-pseudorandom numbers as initialized
		state64.mti = 0
	End Sub


	' Save the PRNG state to a file as XML
	Public Sub saveState64(ByVal fileName As String)
		Try
			Dim serializer As New XmlSerializer(GetType(MTState64))
			Using fs As New FileStream(fileName, FileMode.Create, FileAccess.Write)
				serializer.Serialize(fs, state64)
			End Using
		Catch ex As Exception
			Throw New MTRandom64SaveStateException(ex.Message, fileName)
		End Try
	End Sub


	' Load the PRNG state from a file created by MTRandom64.saveState()
	Public Sub loadState64(ByVal fileName As String)
		Try
			Dim serializer As New XmlSerializer(GetType(MTState64))
			Using fs As New FileStream(fileName, FileMode.Open, FileAccess.Read)
				state64 = DirectCast(serializer.Deserialize(fs), MTState64)
			End Using
		Catch ex As Exception
			Throw New MTRandom64LoadStateException(ex.Message, fileName)
		End Try

	End Sub


	' Initialize constants needed by genrand64_int128SignedInt()
	Private Sub init_int128SignedInt()
#If Need128bitInts Then
		int128SignedIntMaxValue = BigInteger.Parse("170141183460469231731687303715884105727")
		twoToThe128power = BigInteger.Parse("340282366920938463463374607431768211456")
#End If
	End Sub


	' Comparison operator for class MTRandom64: check that two MTRandom64 states ARE equal.
	Public Shared Operator =(ByVal v As MTRandom64, ByVal w As MTRandom64) As Boolean
		' assume the two MTRandom64 states are equal until proven otherwise
		Dim areEqual As Boolean = True

		If v.state64.mti <> w.state64.mti Then
			areEqual = False
		ElseIf v.state64.mt.GetUpperBound(0) <> w.state64.mt.GetUpperBound(0) Then
			areEqual = False
		Else
			Dim i As Integer
			Dim last As Integer = v.state64.mt.GetUpperBound(0)
			For i = 0 To last
				If v.state64.mt(i) <> w.state64.mt(i) Then
					areEqual = False
					Exit For
				End If
			Next i
		End If

		Return areEqual
	End Operator


	' Comparison operator for class MTRandom64: check that two MTRandom64 states ARE NOT equal.
	Public Shared Operator <>(ByVal v As MTRandom64, ByVal w As MTRandom64) As Boolean
		Return Not (v = w)
	End Operator


	' As in mt19937-64.c, initialize with a 64-bit, unsigned seed.
	Public Sub init_genrand64(ByVal seed As ULong)
		Dim tmp As ULong

		state64.mt(0) = seed
		For state64.mti = 1 To (NN - 1)
			tmp = state64.mt(state64.mti - 1)
			state64.mt(state64.mti) = (6364136223846793005UL * (tmp Xor (tmp >> 62)) + CULng(state64.mti))
		Next

	End Sub		'init_genrand64


	' As in mt19937-64.c, initialize with a 64-bit, unsigned array.
	Public Sub init_by_array64(ByRef init_key() As ULong)

		Dim i, j, k As Integer
		Dim key_length As Integer = init_key.Length
		Dim tmp As ULong

		init_genrand64(19650218UL)
		i = 1 : j = 0
		k = CInt(IIf((NN > key_length), NN, key_length))

		For k = k To 1 Step -1
			tmp = state64.mt(i - 1)
			state64.mt(i) = (state64.mt(i) Xor ((tmp Xor (tmp >> 62)) * 3935559000370003845UL)) + init_key(j) + CULng(j)

			i += 1 : j += 1
			If i >= NN Then state64.mt(0) = state64.mt(NN - 1) : i = 1
			If j >= key_length Then j = 0
		Next k

		For k = NN - 1 To 1 Step -1
			tmp = state64.mt(i - 1)
			state64.mt(i) = (state64.mt(i) Xor ((tmp Xor (tmp >> 62)) * 2862933555777941757UL)) - CULng(i)

			i += 1
			If i >= NN Then state64.mt(0) = state64.mt(NN - 1) : i = 1
		Next k

		state64.mt(0) = (1UL << 63)		'/* MSB is 1; assuring non-zero initial array */
	End Sub		'init_by_array64


	' As in mt19937-64.c, generate a 64-bit, unsigned pseudorandom number. 
	Public Function genrand64_int64() As ULong
		Dim x As ULong
		Dim i As Integer

		If (state64.mti >= NN) Then	' /* generate NN words at one time */

			If state64.mti = NN + 1 Then
				init_genrand64(5489UL)
			End If

			For i = 0 To (NN - MM - 1)
				x = (state64.mt(i) And UM) Or (state64.mt(i + 1) And LM)
				state64.mt(i) = state64.mt(i + MM) Xor (x >> 1) Xor ((x And 1UL) * MATRIX_A)
			Next i

			For i = i To (NN - 1 - 1)
				x = (state64.mt(i) And UM) Or (state64.mt(i + 1) And LM)
				state64.mt(i) = state64.mt(i + (MM - NN)) Xor (x >> 1) Xor ((x And 1UL) * MATRIX_A)
			Next i

			x = (state64.mt(NN - 1) And UM) Or (state64.mt(0) And LM)
			state64.mt(NN - 1) = state64.mt(MM - 1) Xor (x >> 1) Xor ((x And 1UL) * MATRIX_A)

			state64.mti = 0
		End If

		x = state64.mt(state64.mti) : state64.mti += 1

		'/* Tempering */
		x = x Xor ((x >> 29) And &H5555555555555555UL)
		x = x Xor ((x << 17) And &H71D67FFFEDA60000UL)
		x = x Xor ((x << 37) And &HFFF7EEE000000000UL)

		Return x Xor (x >> 43)
	End Function	'genrand64_int64


	' As in mt19937-64.c, generate a random number on [0,2^63-1]-interval
	Public Function genrand64_int63() As Long
		Return CLng(genrand64_int64() >> 1)
	End Function	'genrand64_int63


	' As in mt19937-64.c, generate a random number on [0,1]-real-interval
	Public Function genrand64_real1() As Double
		Return (genrand64_int64() >> 11) * (1.0# / 9007199254740991.0#)
	End Function	'genrand64_real1


	' As in mt19937-64.c, generate a random number on [0,1)-real-interval
	Public Function genrand64_real2() As Double
		' This function returns exactly the same values as does the
		' C version.  Any differences in displayed numbers
		' are due to rounding by VB when converting Double to String.

		Return (genrand64_int64() >> 11) * (1.0# / 9007199254740992.0#)
	End Function	'genrand64_real2


	' As in mt19937-64.c, generate a random number on (0,1)-real-interval
	Public Function genrand64_real3() As Double	'double genrand64_real3(void)
		Return ((genrand64_int64() >> 12) + 0.5#) * (1.0# / 4503599627370496.0#)
	End Function	'genrand64_real3


	' The following functions were added by Ron Charlton 2008-09-23 for Visual Basic .NET
	' and revised 2014-02-18 by Ron Charlton for 64 bits.

	' Returns a 64-bit, unsigned integer in [0,n] for n from 0 to 2^64-1
	' Its lowest value is : 0
	' Its highest value is: 18446744073709551615 but less that or equal to n
	Public Function genrand64_intMax(ByVal n As ULong) As ULong

		' Translated by Ron Charlton from C++ file 'MersenneTwister.h' where it is named
		' MTRand::randInt(const uint32& n), and has the following comments:
		'-----
		' Mersenne Twister random number generator -- a C++ class MTRand
		' Based on code by Makoto Matsumoto, Takuji Nishimura, and Shawn Cokus
		' Richard J. Wagner  v1.0  15 May 2003  rjwagner@writeme.com

		' Copyright (C) 1997 - 2002, Makoto Matsumoto and Takuji Nishimura,
		' Copyright (C) 2000 - 2003, Richard J. Wagner
		' All rights reserved.                          
		'-----
		' Wagner's code has the same redistribution requirements as the original
		' Mersenne Twister authors' code.

		' MersenneTwister.h can be found at
		' http://www-personal.umich.edu/~wagnerr/MersenneTwister.html.

		' Find which bits are used in N
		' Optimized by Magnus Jonsson (magnus@smartelectronix.com)
		Dim used As ULong = n
		used = used Or (used >> 1)
		used = used Or (used >> 2)
		used = used Or (used >> 4)
		used = used Or (used >> 8)
		used = used Or (used >> 16)
		used = used Or (used >> 32)

		' Draw numbers until one is found in [0,n]
		Dim i As ULong
		Do
			i = genrand64_int64() And used	' toss unused bits to shorten search
		Loop While i > n

		Return i
	End Function


	' Generate a pseudorandom 64-bit, unsigned integer between lower inclusive and upper inclusive for
	' 0 <= lower <= upper <= 18446744073709551615 (2^64-1).
	Public Function genrand64_intRange(ByVal lower As ULong, ByVal upper As ULong) As ULong
		' Returns a ULong in the range [lower,upper].
		' Its lowest value is : 0 but >= lower
		' Its highest value is: 18446744073709551615 but <= upper
		'
		' Written by Ron Charlton 2008-09-23, and revised 2014-02-18 for 64 bits.
		If lower = upper Then
			Return upper
		End If

		If lower > upper Then
			' swap lower and upper
			Dim temp As ULong = lower
			lower = upper
			upper = temp
		End If

		Return lower + genrand64_intMax(upper - lower)
	End Function


#If Need128bitInts Then
	' Generate a signed, 128-bit BigInteger in the range [-2^127,2^127-1]
	Public Function genrand64_int128SignedInt() As BigInteger

		' Make a 128-bit UNsigned integer in the range [0,2^128-1]
		Dim tmp As BigInteger = CType(genrand64_int64(), BigInteger) _
								Or (CType(genrand64_int64(), BigInteger) << 64)

		' Convert to a signed 128-bit BigInteger.  (The "constants" are
		' initialized automatically in init_int128SignedInt().)
		If tmp > int128SignedIntMaxValue Then
			tmp -= twoToThe128power
		End If

		Return tmp
	End Function
#End If
End Class



' ----------[EXCEPTION CLASSES]----------

' The "unable to save the PRNG's state" exception for MTRandom64.
' Properties:
'	  Message - an error description
<SerializableAttribute> Class MTRandom64SaveStateException
	Inherits Exception

	Private msg As String

	Public Sub New(ByVal MyBaseMessage As String, ByVal fileName As String)
		Const NL As String = ControlChars.NewLine

		msg = "Failed to save the PRN generator state to file" & NL & _
			"""" & fileName & """ in MTRandom64.saveState." & NL & _
			"Possible causes: file does not have write permission, or directory or file name" & NL & _
			"is invalid.  Diagnostic information follows." & NL & _
			NL & MyBaseMessage
	End Sub

	Public Overrides ReadOnly Property Message() As String
		Get
			Return msg
		End Get
	End Property
End Class


' The "unable to load the PRNG's state" exception for MTRandom64.
' Properties:
'	  Message - an error description
<SerializableAttribute> Class MTRandom64LoadStateException
	Inherits Exception

	Private msg As String
	Private stkTrace As String	' stack trace

	Public Sub New(ByVal MyBaseMessage As String, ByVal fileName As String)
		Const NL As String = ControlChars.NewLine

		msg = "Failed to load the PRN generator state from file" & NL & _
			"""" & fileName & """ in MTRandom64.loadState." & NL & _
			"Possible causes: file was not saved by this program or the file does not" & NL & _
			"exist.  Diagnostic information follows." & NL & _
			NL & MyBaseMessage
	End Sub

	Public Overrides ReadOnly Property Message() As String
		Get
			Return msg
		End Get
	End Property
End Class
